@@ -1,55 +1,66 @@
 # -*- coding: utf-8 -*-
+"""Result backend base classes.
+
+- :class:`BaseBackend` defines the interface.
+
+- :class:`KeyValueStoreBackend` is a common base class
+    using K/V semantics like _get and _put.
 """
-    celery.backends.base
-    ~~~~~~~~~~~~~~~~~~~~
+from __future__ import absolute_import, unicode_literals
 
-    Result backend base classes.
-
-    - :class:`BaseBackend` defines the interface.
-
-    - :class:`KeyValueStoreBackend` is a common base class
-      using K/V semantics like _get and _put.
-
-"""
-from __future__ import absolute_import
-
-import time
 import sys
-
+import time
+from collections import namedtuple
 from datetime import timedelta
+from functools import partial
+from weakref import WeakValueDictionary
 
 from billiard.einfo import ExceptionInfo
-from kombu.serialization import (
-    dumps, loads, prepare_accept_content,
-    registry as serializer_registry,
-)
+from kombu.serialization import dumps, loads, prepare_accept_content
+from kombu.serialization import registry as serializer_registry
 from kombu.utils.encoding import bytes_to_str, ensure_bytes, from_utf8
 from kombu.utils.url import maybe_sanitize_url
 
-from celery import states
-from celery import current_app, maybe_signature
-from celery.app import current_task
-from celery.exceptions import ChordError, TimeoutError, TaskRevokedError
-from celery.five import items
-from celery.result import (
-    GroupResult, ResultBase, allow_join_result, result_from_tuple,
-)
-from celery.utils import timeutils
-from celery.utils.functional import LRUCache
+from celery import current_app, group, maybe_signature, states
+from celery._state import get_current_task
+from celery.exceptions import (ChordError, ImproperlyConfigured,
+                               TaskRevokedError, TimeoutError)
+from celery.five import PY3, items
+from celery.result import (GroupResult, ResultBase, allow_join_result,
+                           result_from_tuple)
+from celery.utils.collections import BufferMap
+from celery.utils.functional import LRUCache, arity_greater
 from celery.utils.log import get_logger
-from celery.utils.serialization import (
-    get_pickled_exception,
-    get_pickleable_exception,
-    create_exception_cls,
-)
+from celery.utils.serialization import (create_exception_cls,
+                                        get_pickleable_exception,
+                                        get_pickled_exception)
 
-__all__ = ['BaseBackend', 'KeyValueStoreBackend', 'DisabledBackend']
+__all__ = ('BaseBackend', 'KeyValueStoreBackend', 'DisabledBackend')
 
-EXCEPTION_ABLE_CODECS = frozenset(['pickle'])
-PY3 = sys.version_info >= (3, 0)
+EXCEPTION_ABLE_CODECS = frozenset({'pickle'})
 
 logger = get_logger(__name__)
 
+MESSAGE_BUFFER_MAX = 8192
+
+pending_results_t = namedtuple('pending_results_t', (
+    'concrete', 'weak',
+))
+
+E_NO_BACKEND = """
+No result backend is configured.
+Please see the documentation for more information.
+"""
+
+E_CHORD_NO_BACKEND = """
+Starting chords requires a result backend to be configured.
+
+Note that a group chained with a task is also upgraded to be a chord,
+as this pattern requires synchronization.
+
+Result backends that supports chords: Redis, Database, Memcached, and more.
+"""
+
 
 def unpickle_backend(cls, args, kwargs):
     """Return an unpickled backend."""
@@ -63,7 +74,8 @@ class _nulldict(dict):
     __setitem__ = update = setdefault = ignore
 
 
-class BaseBackend(object):
+class Backend(object):
+
     READY_STATES = states.READY_STATES
     UNREADY_STATES = states.UNREADY_STATES
     EXCEPTION_STATES = states.EXCEPTION_STATES
@@ -79,7 +91,7 @@ class BaseBackend(object):
     supports_native_join = False
 
     #: If true the backend must automatically expire results.
-    #: The daily backend_cleanup periodic task will not be triggered
+    #: The daily backend_cleanup periodic task won't be triggered
     #: in this case.
     supports_autoexpire = False
 
@@ -95,22 +107,25 @@ class BaseBackend(object):
 
     def __init__(self, app,
                  serializer=None, max_cached_results=None, accept=None,
-                 url=None, **kwargs):
+                 expires=None, expires_type=None, url=None, **kwargs):
         self.app = app
         conf = self.app.conf
-        self.serializer = serializer or conf.CELERY_RESULT_SERIALIZER
+        self.serializer = serializer or conf.result_serializer
         (self.content_type,
          self.content_encoding,
          self.encoder) = serializer_registry._encoders[self.serializer]
-        cmax = max_cached_results or conf.CELERY_MAX_CACHED_RESULTS
+        cmax = max_cached_results or conf.result_cache_max
         self._cache = _nulldict() if cmax == -1 else LRUCache(limit=cmax)
+
+        self.expires = self.prepare_expires(expires, expires_type)
         self.accept = prepare_accept_content(
-            conf.CELERY_ACCEPT_CONTENT if accept is None else accept,
-        )
+            conf.accept_content if accept is None else accept)
+        self._pending_results = pending_results_t({}, WeakValueDictionary())
+        self._pending_messages = BufferMap(MESSAGE_BUFFER_MAX)
         self.url = url
 
     def as_uri(self, include_password=False):
-        """Return the backend as an URI, sanitizing the password or not"""
+        """Return the backend as an URI, sanitizing the password or not."""
         # when using maybe_sanitize_url(), "/" is added
         # we're stripping it for consistency
         if include_password:
@@ -119,30 +134,87 @@ class BaseBackend(object):
         return url[:-1] if url.endswith(':///') else url
 
     def mark_as_started(self, task_id, **meta):
-        """Mark a task as started"""
-        return self.store_result(task_id, meta, status=states.STARTED)
+        """Mark a task as started."""
+        return self.store_result(task_id, meta, states.STARTED)
 
-    def mark_as_done(self, task_id, result, request=None):
+    def mark_as_done(self, task_id, result,
+                     request=None, store_result=True, state=states.SUCCESS):
         """Mark task as successfully executed."""
-        return self.store_result(task_id, result,
-                                 status=states.SUCCESS, request=request)
+        if store_result:
+            self.store_result(task_id, result, state, request=request)
+        if request and request.chord:
+            self.on_chord_part_return(request, state, result)
 
-    def mark_as_failure(self, task_id, exc, traceback=None, request=None):
-        """Mark task as executed with failure. Stores the exception."""
-        return self.store_result(task_id, exc, status=states.FAILURE,
+    def mark_as_failure(self, task_id, exc,
+                        traceback=None, request=None,
+                        store_result=True, call_errbacks=True,
+                        state=states.FAILURE):
+        """Mark task as executed with failure."""
+        if store_result:
+            self.store_result(task_id, exc, state,
+                              traceback=traceback, request=request)
+        if request:
+            if request.chord:
+                self.on_chord_part_return(request, state, exc)
+            if call_errbacks and request.errbacks:
+                self._call_task_errbacks(request, exc, traceback)
+
+    def _call_task_errbacks(self, request, exc, traceback):
+        old_signature = []
+        for errback in request.errbacks:
+            errback = self.app.signature(errback)
+            if (
+                # workaround to support tasks with bind=True executed as
+                # link errors. Otherwise retries can't be used
+                not isinstance(errback.type.__header__, partial) and
+                arity_greater(errback.type.__header__, 1)
+            ):
+                errback(request, exc, traceback)
+            else:
+                old_signature.append(errback)
+        if old_signature:
+            # Previously errback was called as a task so we still
+            # need to do so if the errback only takes a single task_id arg.
+            task_id = request.id
+            root_id = request.root_id or task_id
+            group(old_signature, app=self.app).apply_async(
+                (task_id,), parent_id=task_id, root_id=root_id
+            )
+
+    def mark_as_revoked(self, task_id, reason='',
+                        request=None, store_result=True, state=states.REVOKED):
+        exc = TaskRevokedError(reason)
+        if store_result:
+            self.store_result(task_id, exc, state,
+                              traceback=None, request=request)
+        if request and request.chord:
+            self.on_chord_part_return(request, state, exc)
+
+    def mark_as_retry(self, task_id, exc, traceback=None,
+                      request=None, store_result=True, state=states.RETRY):
+        """Mark task as being retries.
+
+        Note:
+            Stores the current exception (if any).
+        """
+        return self.store_result(task_id, exc, state,
                                  traceback=traceback, request=request)
 
     def chord_error_from_stack(self, callback, exc=None):
-        from celery import group
+        # need below import for test for some crazy reason
+        from celery import group  # pylint: disable
         app = self.app
-        backend = app._tasks[callback.task].backend
+        try:
+            backend = app._tasks[callback.task].backend
+        except KeyError:
+            backend = self
         try:
             group(
                 [app.signature(errback)
                  for errback in callback.options.get('link_error') or []],
                 app=app,
-            ).apply_async((callback.id, ))
-        except Exception as eb_exc:
+            ).apply_async((callback.id,))
+        except Exception as eb_exc:  # pylint: disable=broad-except
             return backend.fail_from_current_stack(callback.id, exc=eb_exc)
         else:
             return backend.fail_from_current_stack(callback.id, exc=exc)
@@ -155,32 +227,31 @@ class BaseBackend(object):
             self.mark_as_failure(task_id, exc, ei.traceback)
             return ei
         finally:
-            del(tb)
-
-    def mark_as_retry(self, task_id, exc, traceback=None, request=None):
-        """Mark task as being retries. Stores the current
-        exception (if any)."""
-        return self.store_result(task_id, exc, status=states.RETRY,
-                                 traceback=traceback, request=request)
-
-    def mark_as_revoked(self, task_id, reason='', request=None):
-        return self.store_result(task_id, TaskRevokedError(reason),
-                                 status=states.REVOKED, traceback=None,
-                                 request=request)
+            del tb
 
     def prepare_exception(self, exc, serializer=None):
         """Prepare exception for serialization."""
         serializer = self.serializer if serializer is None else serializer
         if serializer in EXCEPTION_ABLE_CODECS:
             return get_pickleable_exception(exc)
-        return {'exc_type': type(exc).__name__, 'exc_message': str(exc)}
+        return {'exc_type': type(exc).__name__,
+                'exc_message': exc.args,
+                'exc_module': type(exc).__module__}
 
     def exception_to_python(self, exc):
         """Convert serialized exception to Python exception."""
         if exc:
             if not isinstance(exc, BaseException):
-                exc = create_exception_cls(
-                    from_utf8(exc['exc_type']), __name__)(exc['exc_message'])
+                exc_module = exc.get('exc_module')
+                if exc_module is None:
+                    cls = create_exception_cls(
+                        from_utf8(exc['exc_type']), __name__)
+                else:
+                    exc_module = from_utf8(exc_module)
+                    exc_type = from_utf8(exc['exc_type'])
+                    cls = getattr(sys.modules[exc_module], exc_type)
+                exc_msg = exc['exc_message']
+                exc = cls(*exc_msg if isinstance(exc_msg, tuple) else exc_msg)
             if self.serializer in EXCEPTION_ABLE_CODECS:
                 exc = get_pickled_exception(exc)
         return exc
@@ -192,9 +263,12 @@ class BaseBackend(object):
         return result
 
     def encode(self, data):
-        _, _, payload = dumps(data, serializer=self.serializer)
+        _, _, payload = self._encode(data)
         return payload
 
+    def _encode(self, data):
+        return dumps(data, serializer=self.serializer)
+
     def meta_from_decoded(self, meta):
         if meta['status'] in self.EXCEPTION_STATES:
             meta['result'] = self.exception_to_python(meta['result'])
@@ -210,38 +284,11 @@ class BaseBackend(object):
                      content_encoding=self.content_encoding,
                      accept=self.accept)
 
-    def wait_for(self, task_id,
-                 timeout=None, interval=0.5, no_ack=True, on_interval=None):
-        """Wait for task and return its result.
-
-        If the task raises an exception, this exception
-        will be re-raised by :func:`wait_for`.
-
-        If `timeout` is not :const:`None`, this raises the
-        :class:`celery.exceptions.TimeoutError` exception if the operation
-        takes longer than `timeout` seconds.
-
-        """
-
-        time_elapsed = 0.0
-
-        while 1:
-            meta = self.get_task_meta(task_id)
-            if meta['status'] in states.READY_STATES:
-                return meta
-            if on_interval:
-                on_interval()
-            # avoid hammering the CPU checking status.
-            time.sleep(interval)
-            time_elapsed += interval
-            if timeout and time_elapsed >= timeout:
-                raise TimeoutError('The operation timed out.')
-
     def prepare_expires(self, value, type=None):
         if value is None:
-            value = self.app.conf.CELERY_TASK_RESULT_EXPIRES
+            value = self.app.conf.result_expires
         if isinstance(value, timedelta):
-            value = timeutils.timedelta_seconds(value)
+            value = value.total_seconds()
         if value is not None and type:
             return type(value)
         return value
@@ -249,13 +296,11 @@ class BaseBackend(object):
     def prepare_persistent(self, enabled=None):
         if enabled is not None:
             return enabled
-        p = self.app.conf.CELERY_RESULT_PERSISTENT
+        p = self.app.conf.result_persistent
         return self.persistent if p is None else p
 
-    def encode_result(self, result, status):
-        if isinstance(result, ExceptionInfo):
-            result = result.exception
-        if status in self.EXCEPTION_STATES and isinstance(result, Exception):
+    def encode_result(self, result, state):
+        if state in self.EXCEPTION_STATES and isinstance(result, Exception):
             return self.prepare_exception(result)
         else:
             return self.prepare_value(result)
@@ -263,11 +308,11 @@ class BaseBackend(object):
     def is_cached(self, task_id):
         return task_id in self._cache
 
-    def store_result(self, task_id, result, status,
+    def store_result(self, task_id, result, state,
                      traceback=None, request=None, **kwargs):
         """Update task state and result."""
-        result = self.encode_result(result, status)
-        self._store_result(task_id, result, status, traceback,
+        result = self.encode_result(result, state)
+        self._store_result(task_id, result, state, traceback,
                            request=request, **kwargs)
         return result
 
@@ -278,9 +323,10 @@ class BaseBackend(object):
     def _forget(self, task_id):
         raise NotImplementedError('backend does not implement forget.')
 
-    def get_status(self, task_id):
-        """Get the status of a task."""
+    def get_state(self, task_id):
+        """Get the state of a task."""
         return self.get_task_meta(task_id)['status']
+    get_status = get_state  # XXX compat
 
     def get_traceback(self, task_id):
         """Get the traceback for a failed task."""
@@ -297,7 +343,13 @@ class BaseBackend(object):
         except KeyError:
             pass
 
+    def _ensure_not_eager(self):
+        if self.app.conf.task_always_eager:
+            raise RuntimeError(
+                "Cannot retrieve result with task_always_eager enabled")
+
     def get_task_meta(self, task_id, cache=True):
+        self._ensure_not_eager()
         if cache:
             try:
                 return self._cache[task_id]
@@ -318,6 +370,7 @@ class BaseBackend(object):
         self._cache[group_id] = self.get_group_meta(group_id, cache=False)
 
     def get_group_meta(self, group_id, cache=True):
+        self._ensure_not_eager()
         if cache:
             try:
                 return self._cache[group_id]
@@ -344,43 +397,129 @@ class BaseBackend(object):
         return self._delete_group(group_id)
 
     def cleanup(self):
-        """Backend cleanup. Is run by
-        :class:`celery.task.DeleteExpiredTaskMetaTask`."""
-        pass
+        """Backend cleanup.
+
+        Note:
+            This is run by :class:`celery.task.DeleteExpiredTaskMetaTask`.
+        """
 
     def process_cleanup(self):
         """Cleanup actions to do at the end of a task worker process."""
-        pass
 
     def on_task_call(self, producer, task_id):
         return {}
 
-    def on_chord_part_return(self, task, state, result, propagate=False):
+    def add_to_chord(self, chord_id, result):
+        raise NotImplementedError('Backend does not support add_to_chord')
+
+    def on_chord_part_return(self, request, state, result, **kwargs):
         pass
 
-    def fallback_chord_unlock(self, group_id, body, result=None,
-                              countdown=1, **kwargs):
-        kwargs['result'] = [r.as_tuple() for r in result]
+    def fallback_chord_unlock(self, header_result, body, countdown=1,
+                              **kwargs):
+        kwargs['result'] = [r.as_tuple() for r in header_result]
+        queue = body.options.get('queue', getattr(body.type, 'queue', None))
         self.app.tasks['celery.chord_unlock'].apply_async(
-            (group_id, body, ), kwargs, countdown=countdown,
+            (header_result.id, body,), kwargs,
+            countdown=countdown,
+            queue=queue,
         )
 
-    def apply_chord(self, header, partial_args, group_id, body, **options):
-        result = header(*partial_args, task_id=group_id)
-        self.fallback_chord_unlock(group_id, body, **options)
-        return result
+    def ensure_chords_allowed(self):
+        pass
+
+    def apply_chord(self, header_result, body, **kwargs):
+        self.ensure_chords_allowed()
+        self.fallback_chord_unlock(header_result, body, **kwargs)
 
     def current_task_children(self, request=None):
-        request = request or getattr(current_task(), 'request', None)
+        request = request or getattr(get_current_task(), 'request', None)
         if request:
             return [r.as_tuple() for r in getattr(request, 'children', [])]
 
     def __reduce__(self, args=(), kwargs={}):
         return (unpickle_backend, (self.__class__, args, kwargs))
-BaseDictBackend = BaseBackend  # XXX compat
 
 
-class KeyValueStoreBackend(BaseBackend):
+class SyncBackendMixin(object):
+
+    def iter_native(self, result, timeout=None, interval=0.5, no_ack=True,
+                    on_message=None, on_interval=None):
+        self._ensure_not_eager()
+        results = result.results
+        if not results:
+            return iter([])
+        return self.get_many(
+            {r.id for r in results},
+            timeout=timeout, interval=interval, no_ack=no_ack,
+            on_message=on_message, on_interval=on_interval,
+        )
+
+    def wait_for_pending(self, result, timeout=None, interval=0.5,
+                         no_ack=True, on_message=None, on_interval=None,
+                         callback=None, propagate=True):
+        self._ensure_not_eager()
+        if on_message is not None:
+            raise ImproperlyConfigured(
+                'Backend does not support on_message callback')
+
+        meta = self.wait_for(
+            result.id, timeout=timeout,
+            interval=interval,
+            on_interval=on_interval,
+            no_ack=no_ack,
+        )
+        if meta:
+            result._maybe_set_cache(meta)
+            return result.maybe_throw(propagate=propagate, callback=callback)
+
+    def wait_for(self, task_id,
+                 timeout=None, interval=0.5, no_ack=True, on_interval=None):
+        """Wait for task and return its result.
+
+        If the task raises an exception, this exception
+        will be re-raised by :func:`wait_for`.
+
+        Raises:
+            celery.exceptions.TimeoutError:
+                If `timeout` is not :const:`None`, and the operation
+                takes longer than `timeout` seconds.
+        """
+        self._ensure_not_eager()
+
+        time_elapsed = 0.0
+
+        while 1:
+            meta = self.get_task_meta(task_id)
+            if meta['status'] in states.READY_STATES:
+                return meta
+            if on_interval:
+                on_interval()
+            # avoid hammering the CPU checking status.
+            time.sleep(interval)
+            time_elapsed += interval
+            if timeout and time_elapsed >= timeout:
+                raise TimeoutError('The operation timed out.')
+
+    def add_pending_result(self, result, weak=False):
+        return result
+
+    def remove_pending_result(self, result):
+        return result
+
+    @property
+    def is_async(self):
+        return False
+
+
+class BaseBackend(Backend, SyncBackendMixin):
+    """Base (synchronous) result backend."""
+
+
+BaseDictBackend = BaseBackend  # noqa: E305 XXX compat
+
+
+class BaseKeyValueStoreBackend(Backend):
     key_t = ensure_bytes
     task_keyprefix = 'celery-task-meta-'
     group_keyprefix = 'celery-taskset-meta-'
@@ -388,10 +527,10 @@ class KeyValueStoreBackend(BaseBackend):
     implements_incr = False
 
     def __init__(self, *args, **kwargs):
-        if hasattr(self.key_t, '__func__'):
+        if hasattr(self.key_t, '__func__'):  # pragma: no cover
             self.key_t = self.key_t.__func__  # remove binding
         self._encode_prefixes()
-        super(KeyValueStoreBackend, self).__init__(*args, **kwargs)
+        super(BaseKeyValueStoreBackend, self).__init__(*args, **kwargs)
         if self.implements_incr:
             self.apply_chord = self._apply_chord_incr
 
@@ -440,7 +579,7 @@ class KeyValueStoreBackend(BaseBackend):
         ])
 
     def _strip_prefix(self, key):
-        """Takes bytes, emits string."""
+        """Take bytes: emit string."""
         key = self.key_t(key)
         for prefix in self.task_keyprefix, self.group_keyprefix:
             if key.startswith(prefix):
@@ -457,14 +596,19 @@ class KeyValueStoreBackend(BaseBackend):
     def _mget_to_results(self, values, keys):
         if hasattr(values, 'items'):
             # client returns dict so mapping preserved.
-            return dict((self._strip_prefix(k), v)
-                        for k, v in self._filter_ready(items(values)))
+            return {
+                self._strip_prefix(k): v
+                for k, v in self._filter_ready(items(values))
+            }
         else:
             # client returns list so need to recreate mapping.
-            return dict((bytes_to_str(keys[i]), v)
-                        for i, v in self._filter_ready(enumerate(values)))
+            return {
+                bytes_to_str(keys[i]): v
+                for i, v in self._filter_ready(enumerate(values))
+            }
 
     def get_many(self, task_ids, timeout=None, interval=0.5, no_ack=True,
+                 on_message=None, on_interval=None, max_iterations=None,
                  READY_STATES=states.READY_STATES):
         interval = 0.5 if interval is None else interval
         ids = task_ids if isinstance(task_ids, set) else set(task_ids)
@@ -487,21 +631,30 @@ class KeyValueStoreBackend(BaseBackend):
             r = self._mget_to_results(self.mget([self.get_key_for_task(k)
                                                  for k in keys]), keys)
             cache.update(r)
-            ids.difference_update(set(bytes_to_str(v) for v in r))
+            ids.difference_update({bytes_to_str(v) for v in r})
             for key, value in items(r):
+                if on_message is not None:
+                    on_message(value)
                 yield bytes_to_str(key), value
             if timeout and iterations * interval >= timeout:
                 raise TimeoutError('Operation timed out ({0})'.format(timeout))
+            if on_interval:
+                on_interval()
             time.sleep(interval)  # don't busy loop.
             iterations += 1
+            if max_iterations and iterations >= max_iterations:
+                break
 
     def _forget(self, task_id):
         self.delete(self.get_key_for_task(task_id))
 
-    def _store_result(self, task_id, result, status,
+    def _store_result(self, task_id, result, state,
                       traceback=None, request=None, **kwargs):
-        meta = {'status': status, 'result': result, 'traceback': traceback,
-                'children': self.current_task_children(request)}
+        meta = {
+            'status': state, 'result': result, 'traceback': traceback,
+            'children': self.current_task_children(request),
+            'task_id': bytes_to_str(task_id),
+        }
         self.set(self.get_key_for_task(task_id), self.encode(meta))
         return result
 
@@ -514,14 +667,14 @@ class KeyValueStoreBackend(BaseBackend):
         self.delete(self.get_key_for_group(group_id))
 
     def _get_task_meta_for(self, task_id):
-        """Get task metadata for a task by id."""
+        """Get task meta-data for a task by id."""
         meta = self.get(self.get_key_for_task(task_id))
         if not meta:
             return {'status': states.PENDING, 'result': None}
         return self.decode_result(meta)
 
     def _restore_group(self, group_id):
-        """Get task metadata for a task by id."""
+        """Get task meta-data for a task by id."""
         meta = self.get(self.get_key_for_group(group_id))
         # previously this was always pickled, but later this
         # was extended to support other serializers, so the
@@ -532,26 +685,23 @@ class KeyValueStoreBackend(BaseBackend):
             meta['result'] = result_from_tuple(result, self.app)
             return meta
 
-    def _apply_chord_incr(self, header, partial_args, group_id, body,
-                          result=None, **options):
-        self.save_group(group_id, self.app.GroupResult(group_id, result))
-        return header(*partial_args, task_id=group_id)
+    def _apply_chord_incr(self, header_result, body, **kwargs):
+        self.ensure_chords_allowed()
+        header_result.save(backend=self)
 
-    def on_chord_part_return(self, task, state, result, propagate=None):
+    def on_chord_part_return(self, request, state, result, **kwargs):
         if not self.implements_incr:
             return
         app = self.app
-        if propagate is None:
-            propagate = app.conf.CELERY_CHORD_PROPAGATES
-        gid = task.request.group
+        gid = request.group
         if not gid:
             return
         key = self.get_key_for_chord(gid)
         try:
-            deps = GroupResult.restore(gid, backend=task.backend)
-        except Exception as exc:
-            callback = maybe_signature(task.request.chord, app=app)
-            logger.error('Chord %r raised: %r', gid, exc, exc_info=1)
+            deps = GroupResult.restore(gid, backend=self)
+        except Exception as exc:  # pylint: disable=broad-except
+            callback = maybe_signature(request.chord, app=app)
+            logger.exception('Chord %r raised: %r', gid, exc)
             return self.chord_error_from_stack(
                 callback,
                 ChordError('Cannot restore group: {0!r}'.format(exc)),
@@ -560,25 +710,24 @@ class KeyValueStoreBackend(BaseBackend):
             try:
                 raise ValueError(gid)
             except ValueError as exc:
-                callback = maybe_signature(task.request.chord, app=app)
-                logger.error('Chord callback %r raised: %r', gid, exc,
-                             exc_info=1)
+                callback = maybe_signature(request.chord, app=app)
+                logger.exception('Chord callback %r raised: %r', gid, exc)
                 return self.chord_error_from_stack(
                     callback,
                     ChordError('GroupResult {0} no longer exists'.format(gid)),
                 )
         val = self.incr(key)
         size = len(deps)
-        if val > size:
+        if val > size:  # pragma: no cover
             logger.warning('Chord counter incremented too many times for %r',
                            gid)
         elif val == size:
-            callback = maybe_signature(task.request.chord, app=app)
+            callback = maybe_signature(request.chord, app=app)
             j = deps.join_native if deps.supports_native_join else deps.join
             try:
                 with allow_join_result():
-                    ret = j(timeout=3.0, propagate=propagate)
-            except Exception as exc:
+                    ret = j(timeout=3.0, propagate=True)
+            except Exception as exc:  # pylint: disable=broad-except
                 try:
                     culprit = next(deps._failed_join_report())
                     reason = 'Dependency {0.id} raised {1!r}'.format(
@@ -587,13 +736,13 @@ class KeyValueStoreBackend(BaseBackend):
                 except StopIteration:
                     reason = repr(exc)
 
-                logger.error('Chord %r raised: %r', gid, reason, exc_info=1)
+                logger.exception('Chord %r raised: %r', gid, reason)
                 self.chord_error_from_stack(callback, ChordError(reason))
             else:
                 try:
                     callback.delay(ret)
-                except Exception as exc:
-                    logger.error('Chord %r raised: %r', gid, exc, exc_info=1)
+                except Exception as exc:  # pylint: disable=broad-except
+                    logger.exception('Chord %r raised: %r', gid, exc)
                     self.chord_error_from_stack(
                         callback,
                         ChordError('Callback error: {0!r}'.format(exc)),
@@ -602,22 +751,29 @@ class KeyValueStoreBackend(BaseBackend):
                 deps.delete()
                 self.client.delete(key)
         else:
-            self.expire(key, 86400)
+            self.expire(key, self.expires)
+
+
+class KeyValueStoreBackend(BaseKeyValueStoreBackend, SyncBackendMixin):
+    """Result backend base class for key/value stores."""
 
 
 class DisabledBackend(BaseBackend):
+    """Dummy result backend."""
+
     _cache = {}   # need this attribute to reset cache in tests.
 
     def store_result(self, *args, **kwargs):
         pass
 
+    def ensure_chords_allowed(self):
+        raise NotImplementedError(E_CHORD_NO_BACKEND.strip())
+
     def _is_disabled(self, *args, **kwargs):
-        raise NotImplementedError(
-            'No result backend configured.  '
-            'Please see the documentation for more information.')
+        raise NotImplementedError(E_NO_BACKEND.strip())
 
     def as_uri(self, *args, **kwargs):
         return 'disabled://'
 
     get_state = get_status = get_result = get_traceback = _is_disabled
-    wait_for = get_many = _is_disabled
+    get_task_meta_for = wait_for = get_many = _is_disabled
